package gov.usgs.earthquake.nshmp.data;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.data.DoubleData.areMonotonic;

import java.util.Arrays;
import java.util.Collection;
import java.util.TreeMap;
import java.util.function.Predicate;
import java.util.stream.Collectors;

/**
 * Static methods we don't want in public interfaces.
 * 
 * @author U.S. Geological Survey
 */
class Sequences {

  private Sequences() {}

  /* Assumes array copies have already been created. */
  static XySequence construct(
      double[] xs,
      double[] ys) {
    validateArrays(xs, ys);
    return new ArrayXySequence(xs, ys);
  }

  /* Assumes array copies have already been created. */
  static MutableXySequence constructMutable(
      double[] xs,
      double[] ys) {
    validateArrays(xs, ys);
    return new MutableArrayXySequence(xs, ys);
  }

  static void validateArrays(double[] xs, double[] ys) {
    checkArgument(xs.length > 0, "x-values may not be empty");
    checkArgument(xs.length == ys.length, "x- and y-values are different sizes");
    checkArgument(areMonotonic(true, true, xs), "x-values do not increase monotonically");
  }

  static XySequence trim(ArrayXySequence sequence, boolean mutable) {
    checkState(!sequence.isClear(), "XySequence.trim() not permitted for 'clear' sequences");
    int minIndex = DoubleData.firstNonZeroIndex(sequence.ys);
    int maxIndex = DoubleData.lastNonZeroIndex(sequence.ys) + 1;
    double[] xs = Arrays.copyOfRange(sequence.xs, minIndex, maxIndex);
    double[] ys = Arrays.copyOfRange(sequence.ys, minIndex, maxIndex);
    return mutable ? new MutableArrayXySequence(xs, ys) : new ArrayXySequence(xs, ys);
  }

  /**
   * Returns {@code true} if all {@code XySequence} in {@code Collection} have
   * the same x-values. Returns {@code true} if {@code Collection} size is 1.
   * 
   * @throws IllegalStateException if {@code Collection<XySequence>} is empty
   * 
   * @param xys {@code Collection<XySequence>} to be tested for similarity
   * 
   * @return {@code true} if all {@code XySequence}s in specified collection
   *         share the same x-values
   */
  static boolean areSimilar(Collection<XySequence> xys) {
    checkState(!xys.isEmpty());
    if (xys.size() == 1) {
      return true;
    }
    // Safe covariant cast
    ArrayXySequence firstXy = (ArrayXySequence) xys.iterator().next();
    Predicate<XySequence> firstXsPredicate = other -> {
      return Arrays.equals(
          firstXy.xs,
          ((ArrayXySequence) other).xs);
    };
    return xys.stream()
        .skip(1)
        .allMatch(firstXsPredicate);
  }

  /**
   * Combine {@code Collection} of {@code XySequence} such that the resulting
   * x-values are the union of all specified sequences, and the y-values are
   * summed accordingly.
   * 
   * @param xys to combine
   * @return xys
   */
  static XySequence combine(Collection<XySequence> xys) {
    checkState(!xys.isEmpty());
    // Safe covariant cast
    XySequence first = (ArrayXySequence) xys.iterator().next();
    if (xys.size() == 1) {
      return first;
    }
    if (areSimilar(xys)) {
      MutableXySequence result = MutableXySequence.emptyCopyOf(first);
      xys.forEach(result::add);
      return result;
    }

    var combinedMap = xys.stream()
        .flatMap(XySequence::stream)
        .collect(Collectors.toMap(
            XyPoint::x,
            XyPoint::y,
            Double::sum,
            TreeMap::new));
    return XySequence.create(
        combinedMap.keySet(),
        combinedMap.values());
  }

}
