package gov.usgs.earthquake.nshmp.eq.model;

import static com.google.common.base.Preconditions.checkState;

import java.io.BufferedReader;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Range;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;

import gov.usgs.earthquake.nshmp.eq.fault.FocalMech;
import gov.usgs.earthquake.nshmp.eq.model.MfdHelper2.MfdProperties;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.UncertaintyModel;
import gov.usgs.earthquake.nshmp.internal.NshmFaultSection;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/*
 * Wrapper class for source model deserialization methods and helpers.
 * 
 * Developer notes:
 * 
 * Logic trees (*-tree) and other NSHM source model maps (*-map) are difficult
 * to deserialize using a Gson instance due to the use of generics and the
 * complexity of some nested objects. Some objects have conditional fields that
 * may or may not be present (e.g. MFDs), or that may be represented by either
 * an object or a string pointer to an object in a map. Gson.fromJson() methods
 * require the use of TypeTokens, which are difficult to create for generic
 * collections of objects that themselves include parameterized types. A
 * TypteToken based approach also requires either polluting the 'tree' package
 * with types outside the package, or exposing the various branch flavors, which
 * we'd like to avoid. We therefore opt to deserialize most *-tree and *-map
 * objects using a JsonParser.
 */
class Deserialize {

  private static final Gson GSON = new GsonBuilder().create();

  private static final String ID_KEY = "id";
  private static final String WEIGHT_KEY = "weight";
  private static final String VALUE_KEY = "value";
  private static final String TYPE_KEY = "type";
  private static final String SECTIONS_KEY = "sections";
  private static final String MFD_TREE_KEY = "mfd-tree";
  private static final String M_MIN_KEY = "mMin";
  private static final String M_MAX_KEY = "mMax";
  private static final String DEPTH_TREE_KEY = "depth-tree";

  /* Read model info. */
  static HazardModel.Info modelInfo(Path json) throws IOException {
    HazardModel.Info info = fromJson(json, HazardModel.Info.class);
    checkState(info.name != null, "Missing 'name' in %s", json.getFileName());
    return info;
  }

  /* Read a MFD configuration. */
  static MfdConfig mfdConfig(Path json) throws IOException {
    JsonObject config = jsonObject(json);

    JsonElement mMinElem = config.get(MfdConfig.MIN_MAGNITUDE_KEY);
    Double mMin = mMinElem.isJsonNull() ? null : mMinElem.getAsDouble();

    JsonElement epiElem = config.get(MfdConfig.EPISTEMIC_TREE_KEY);
    LogicTree<Double> epiTree = epiElem.isJsonNull() ? null : doubleTree(epiElem);

    JsonElement aleaElem = config.get(MfdConfig.ALEATORY_PROPS_KEY);
    MfdConfig.AleatoryProperties aleaProps = aleaElem.isJsonNull()
        ? null
        : GSON.fromJson(aleaElem, MfdConfig.AleatoryProperties.class);

    return MfdConfig.builder()
        .epistemicTree(epiTree)
        .aleatoryProperties(aleaProps)
        .minimumMagnitude(mMin)
        .build();
  }

  /* Read a GMM configuration. */
  static GmmConfig gmmConfig(Path json) throws IOException {
    JsonObject config = jsonObject(json);

    double rMax = config.get(GmmConfig.MAX_DISTANCE_KEY).getAsDouble();

    JsonElement treeElem = config.get(GmmConfig.EPISTEMIC_TREE_KEY);
    LogicTree<Double> epiTree = treeElem.isJsonNull()
        ? null
        : doubleTree(treeElem);

    JsonElement modelElem = config.get(GmmConfig.EPISTEMIC_MODEL_KEY);
    UncertaintyModel epiModel = modelElem.isJsonNull()
        ? null
        : UncertaintyModel.valueOf(modelElem.getAsString());

    return GmmConfig.builder()
        .maximumDistance(rMax)
        .epistemicTree(epiTree)
        .epistemicModel(epiModel)
        .build();
  }

  /* Create a GMM logic tree. */
  static LogicTree<Gmm> gmmTree(Path json) throws IOException {
    return enumValueTree(json, Gmm.class);
  }

  /* Create a source path logic tree. */
  static LogicTree<Path> sourceTree(Path json) throws IOException {
    LogicTree.Builder<Path> builder = LogicTree.builder();
    for (JsonElement branch : jsonArray(json)) {
      JsonObject obj = branch.getAsJsonObject();
      String id = obj.get(ID_KEY).getAsString();
      builder.addBranch(
          id,
          json.resolveSibling(id),
          obj.get(WEIGHT_KEY).getAsDouble());
    }
    return builder.build();
  }

  /* Create a rate logic tree. */
  static LogicTree<Double> rateTree(Path json) throws IOException {
    return doubleTree(jsonArray(json));
  }

  /* Create an id:section map from a fault-sections directory. */
  static Map<Integer, NshmFaultSection> faultSections(Path dir) throws IOException {
    ImmutableMap.Builder<Integer, NshmFaultSection> sectionMap = ImmutableMap.builder();
    for (Path json : Files.newDirectoryStream(dir, "*.geojson")) {
      NshmFaultSection section = NshmFaultSection.fromGeoJson(json);
      sectionMap.put(section.id(), section);
    }
    return sectionMap.build();
  }

  /* Create a ruptureSet. */
  // static NshmRuptureSet ruptureSet(
  // Path json,
  // Optional<Map<String, LogicTree<MfdProperties>>> mfdMap) throws IOException
  // {
  //
  // return ruptureSet(jsonObject(json), mfdMap);
  // }

  /* Create a string:ruptureSet map from a rupture-map. */
  // @Deprecated TODO clean
  // static Map<String, NshmRuptureSet> ruptureMap(Path json) throws IOException
  // {
  // ImmutableMap.Builder<String, NshmRuptureSet> ruptureMap =
  // ImmutableMap.builder();
  // for (Entry<String, JsonElement> entry : jsonObject(json).entrySet()) {
  // ruptureMap.put(
  // entry.getKey(),
  // ruptureSet(entry.getValue(), null));
  // }
  // return ruptureMap.build();
  // }

  // TODO test that mfdMap.get() throws NoSuchElement when mfd map missing but
  // required by mfd-tree:string

  /* Create a rupture set. */
  static RuptureSet ruptureSet(
      Path json,
      Optional<Map<String, LogicTree<MfdProperties>>> mfdMap) throws IOException {

    // by the time a ruptureSet is encountered all supporting data should exist
    // - optional rate tree
    // - mfdConfig, possibly empty()
    // - ??

    JsonObject obj = jsonObject(json);

    int id = obj.get(ID_KEY).getAsInt();

    /* 'sections' optional; if absent id = section-id. */
    int[] sections = obj.has(SECTIONS_KEY)
        ? GSON.fromJson(obj.get(SECTIONS_KEY), int[].class)
        : new int[] { id };

    JsonElement mfdElement = obj.get(MFD_TREE_KEY);
    LogicTree<MfdProperties> mfdTree = mfdElement.isJsonPrimitive()
        ? mfdMap.get().get(mfdElement.getAsString())
        : mfdPropertiesTree(mfdElement);

    return new RuptureSet(id, sections, mfdTree, null, null); // TODO needs
                                                              // MfdConfig
  }

  /* Create a string:mfd-tree map from a mfd-map. */
  static Map<String, LogicTree<MfdProperties>> mfdMap(Path json) throws IOException {
    ImmutableMap.Builder<String, LogicTree<MfdProperties>> mfdMap = ImmutableMap.builder();
    for (Entry<String, JsonElement> entry : jsonObject(json).entrySet()) {
      mfdMap.put(entry.getKey(), mfdPropertiesTree(entry.getValue()));
    }
    return mfdMap.build();
  }

  /* Create a logic tree of MFD properties. */
  private static LogicTree<MfdProperties> mfdPropertiesTree(JsonElement json) {
    LogicTree.Builder<MfdProperties> builder = LogicTree.builder();
    for (JsonElement branch : json.getAsJsonArray()) {
      JsonObject obj = branch.getAsJsonObject();
      builder.addBranch(
          obj.get(ID_KEY).getAsString(),
          mfdProperties(obj.get(VALUE_KEY)),
          obj.get(WEIGHT_KEY).getAsDouble());
    }
    return builder.build();
  }

  private static MfdProperties mfdProperties(JsonElement json) {
    JsonObject obj = json.getAsJsonObject();
    Mfd.Type type = Mfd.Type.valueOf(obj.get(TYPE_KEY).getAsString());
    JsonElement data = obj.get(VALUE_KEY);
    Object properties;
    switch (type) {
      case SINGLE:
        properties = GSON.fromJson(data, MfdHelper2.Single.class);
        break;
      case GR:
        properties = GSON.fromJson(data, MfdHelper2.GutenbergRichter.class);
        break;
      case GR_TAPER:
        properties = GSON.fromJson(data, MfdHelper2.GrTaper.class);
        break;
      case INCR:
        properties = GSON.fromJson(data, MfdHelper2.Incremental.class);
        break;
      default:
        throw new IllegalArgumentException("Unknown MFD type: " + type);
    }
    return new MfdProperties(type, properties);
  }

  /* Create a MFD logic tree. */
  private static LogicTree<Mfd> mfdTree(JsonElement json) {
    LogicTree.Builder<Mfd> builder = LogicTree.builder();
    for (JsonElement branch : json.getAsJsonArray()) {
      JsonObject obj = branch.getAsJsonObject();
      builder.addBranch(
          obj.get(ID_KEY).getAsString(),
          mfd(obj.get(VALUE_KEY)),
          obj.get(WEIGHT_KEY).getAsDouble());
    }
    return builder.build();
  }

  /* Create a MFD. */
  private static Mfd mfd(JsonElement json) {

    // assume source-config

    JsonObject obj = json.getAsJsonObject();
    Mfd.Type type = Mfd.Type.valueOf(obj.get(TYPE_KEY).getAsString());
    JsonElement data = obj.get(VALUE_KEY);
    switch (type) {
      case SINGLE:
        // LogicTree<Mfd> single = MfdHelper2.createSingle(data);
        MfdHelper2.Single single = GSON.fromJson(data, MfdHelper2.Single.class);
        return null;
      case GR:
        MfdHelper2.GutenbergRichter gr = GSON.fromJson(data, MfdHelper2.GutenbergRichter.class);
        return null;
      case GR_TAPER:
        MfdHelper2.GrTaper grTaper = GSON.fromJson(data, MfdHelper2.GrTaper.class);
        return null;
      case INCR:
        MfdHelper2.Incremental incr = GSON.fromJson(data, MfdHelper2.Incremental.class);
        return null;
      default:
        throw new IllegalArgumentException("Unknown MFD type: " + type);
    }
  }

  /* Create a gridded seismicity source depth model map. */
  static Map<String, GridDepthModel> gridDepthMap(Path json) throws IOException {
    ImmutableMap.Builder<String, GridDepthModel> depthMap = ImmutableMap.builder();
    for (Entry<String, JsonElement> entry : jsonObject(json).entrySet()) {
      depthMap.put(entry.getKey(), depthModel(entry.getValue()));
    }
    return depthMap.build();
  }

  /* Create a gridded seismicity source depth model. */
  private static GridDepthModel depthModel(JsonElement json) {
    JsonObject obj = json.getAsJsonObject();
    Range<Double> mRange = Range.closedOpen(
        obj.get(M_MIN_KEY).getAsDouble(),
        obj.get(M_MAX_KEY).getAsDouble());
    LogicTree<Double> depthTree = doubleTree(obj.get(DEPTH_TREE_KEY));
    return new GridDepthModel(mRange, depthTree);
  }

  /* Create a focal mechanism logic tree. */
  static LogicTree<FocalMech> focalMechTree(Path json) throws IOException {
    return enumValueTree(json, FocalMech.class);
  }

  /* Create an enum valued tree. */
  private static <E extends Enum<E>> LogicTree<E> enumValueTree(Path json, Class<E> clazz)
      throws IOException {
    LogicTree.EnumValueBuilder<E> builder = LogicTree.enumValueBuilder();
    for (JsonElement branch : jsonArray(json)) {
      JsonObject obj = branch.getAsJsonObject();
      builder.addBranch(
          Enum.valueOf(clazz, obj.get(ID_KEY).getAsString()),
          obj.get(WEIGHT_KEY).getAsDouble());
    }
    return builder.build();
  }

  /* Create double valued logic tree. */
  private static LogicTree<Double> doubleTree(JsonElement json) {
    LogicTree.Builder<Double> builder = LogicTree.builder();
    for (JsonElement branch : json.getAsJsonArray()) {
      JsonObject obj = branch.getAsJsonObject();
      builder.addBranch(
          obj.get(ID_KEY).getAsString(),
          obj.get(VALUE_KEY).getAsDouble(),
          obj.get(WEIGHT_KEY).getAsDouble());
    }
    return builder.build();
  }

  /* Buffered reader wrapper for JSON array. */
  private static JsonArray jsonArray(Path json) throws IOException {
    try (BufferedReader br = Files.newBufferedReader(json)) {
      return JsonParser.parseReader(br).getAsJsonArray();
    }
  }

  /* Buffered reader wrapper for JSON object. */
  private static JsonObject jsonObject(Path json) throws IOException {
    try (BufferedReader br = Files.newBufferedReader(json)) {
      return JsonParser.parseReader(br).getAsJsonObject();
    }
  }

  /* Buffered reader wrapper for class. */
  private static <T> T fromJson(Path json, Class<T> clazz) throws IOException {
    try (BufferedReader br = Files.newBufferedReader(json)) {
      return GSON.fromJson(br, clazz);
    }
  }

}
