package gov.usgs.earthquake.nshmp.eq.model;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;

import java.util.Optional;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.data.DoubleData;
import gov.usgs.earthquake.nshmp.gmm.UncertaintyModel;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/*
 * Ground motion model (GMM) configuration.
 * 
 * @author U.S. Geological Survey
 */
class GmmConfig {

  static final Range<Double> MAX_DIST_RANGE = Range.closed(50.0, 1000.0);
  static final String MAX_DISTANCE_KEY = "max-distance";
  static final String EPISTEMIC_MODEL_KEY = "epistemic-model";
  static final String EPISTEMIC_TREE_KEY = "epistemic-tree";

  final double maxDistance;
  final Optional<UncertaintyModel> epistemicModel;
  final Optional<LogicTree<Double>> epistemicTree;

  private GmmConfig(Builder builder) {
    this.maxDistance = builder.maxDistance;
    this.epistemicModel = builder.epistemicModel;
    this.epistemicTree = builder.epistemicTree;
  }

  static Builder builder() {
    return new Builder();
  }

  /* Single use builder. */
  static class Builder {

    private Double maxDistance;
    private Optional<UncertaintyModel> epistemicModel = Optional.empty();
    private Optional<LogicTree<Double>> epistemicTree = Optional.empty();
    private boolean built = false;

    Builder maximumDistance(double rMax) {
      DoubleData.checkInRange(MAX_DIST_RANGE, MAX_DISTANCE_KEY, rMax);
      this.maxDistance = rMax;
      return this;
    }

    Builder epistemicModel(UncertaintyModel model) {
      this.epistemicModel = Optional.ofNullable(model);
      return this;
    }

    Builder epistemicTree(LogicTree<Double> tree) {
      this.epistemicTree = Optional.ofNullable(tree);
      return this;
    }

    GmmConfig build() {
      checkState(!built, "GmmConfig.Builder has already been used");
      checkNotNull(maxDistance, "%s required", MAX_DISTANCE_KEY);
      if (epistemicTree.isPresent()) {
        checkState(epistemicModel.isPresent(), "%s required", EPISTEMIC_MODEL_KEY);
      }
      built = true;
      return new GmmConfig(this);
    }
  }
}
