package gov.usgs.earthquake.nshmp.eq.model;

import static com.google.common.base.CaseFormat.LOWER_HYPHEN;
import static com.google.common.base.CaseFormat.UPPER_UNDERSCORE;
import static com.google.common.base.Preconditions.checkState;
import static java.nio.file.Files.newDirectoryStream;

import java.io.IOException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.NoSuchFileException;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.google.common.collect.Lists;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.reflect.TypeToken;

import gov.usgs.earthquake.nshmp.calc.CalcConfig;
import gov.usgs.earthquake.nshmp.eq.TectonicSetting;
import gov.usgs.earthquake.nshmp.eq.model.HazardModel.Builder;
import gov.usgs.earthquake.nshmp.eq.model.MfdHelper2.MfdProperties;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.internal.NshmFaultSection;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/*
 * Hazard Model loader.
 * 
 * @author U.S. Geological Survey
 */
class Loader {

  static final String MODEL_INFO = "model-info.json";
  static final String SOURCE_CONFIG = SourceConfig.FILE_NAME;
  static final String CALC_CONFIG = CalcConfig.FILE_NAME;

  static final String GMM_TREE = "gmm-tree.json";
  static final String GMM_CONFIG = "gmm-config.json";
  static final String MFD_CONFIG = "mfd-config.json";
  static final String MFD_MAP = "mfd-map.json";
  static final String RATE_TREE = "rate-tree.json";
  static final String RUPTURE_SET = "rupture-set.json";
  static final String SECTION_DIR = "fault-sections";
  static final String SOURCE_TREE = "source-tree.json";

  static final TypeToken<LogicTree<Mfd>> mfdTreeToken = new TypeToken<LogicTree<Mfd>>() {};

  static final Gson GSON = new GsonBuilder()
      // .registerTypeAdapter(NshmRuptureSet.class, new
      // RuptureSetDeserializer())
      // .registerTypeAdapter(mfdTreeToken.getType(), new MfdTreeDeserializer())
      .setPrettyPrinting()
      .create();

  private static Logger log;

  static {
    log = Logger.getLogger(Loader.class.getName());
  }

  /*
   * Load a hazard model. Supplied path should be an absolute path to a
   * directory containing model files as described in the source model
   * documentation.
   * 
   * MFD here generally refers to a MfdProperties object, which can later be
   * used to build individual MFDs
   * 
   * TODO Thread safety?
   */
  static HazardModel load(Path path) throws IOException {

    try {

      HazardModel.Builder builder = HazardModel.builder();

      /* Load required model info. */
      Path infoPath = path.resolve(MODEL_INFO);
      HazardModel.Info info = Deserialize.modelInfo(infoPath);
      // log.info("Model: " + info.name);
      System.out.println("--> " + info.name);
      builder.info(info);

      /* Load required source config. */
      Path srcConfigPath = path.resolve(SOURCE_CONFIG);
      SourceConfig sourceConfig = SourceConfig.fromFile(srcConfigPath).build();
      System.out.println("--> " + sourceConfig.toString());
      // log.info(sourceConfig.toString());

      /* Look for optional calc config. */
      CalcConfig.Builder calcConfigBuilder = CalcConfig.withDefaults();
      Path customCalcConfig = path.resolve(CALC_CONFIG);
      if (Files.exists(customCalcConfig)) {
        calcConfigBuilder.extend(CalcConfig.fromFile(customCalcConfig));
      }
      builder.config(calcConfigBuilder.build());

      /* Process tectonic setting source directories. */
      List<Path> sourceDirList = sourceDirectoryList(path);
      for (Path sourceDir : sourceDirList) {
        loadSourceDirectory(sourceDir, builder, sourceConfig, Optional.empty());
      }

      return builder.build();

      // } catch (IOException ioe) {
      // log.severe("Error loading model: " + path);
      // throw ioe;
    } catch (Exception e) {
      String pathString = (path != null)
          ? path.toAbsolutePath().normalize().toString()
          : "null";
      log.log(Level.SEVERE, "Error loading model: " + pathString, e);
      // log.severe("Error loading model: " + path);
      throw e;
    }
  }

  /* Collect tectonic setting source directories. */
  private static List<Path> sourceDirectoryList(Path path) throws IOException {
    try (DirectoryStream<Path> entries = newDirectoryStream(
        path,
        TectonicSettingFilter.INSTANCE)) {
      List<Path> dirs = Lists.newArrayList(entries);
      checkState(dirs.size() > 0, "No source directories (e.g. active-crust/)");
      return dirs;
    }
  }

  /*
   * Process a source directory. This method is called recursively as a
   * directory is processed. For top-level, tectonic setting directories, 'gmms'
   * will be empty; for nested directories it will be present and may be
   * overridden. Likewise, the supplied model config may also be overriden in
   * nested directories.
   */
  private static void loadSourceDirectory(
      Path dir,
      Builder builder,
      SourceConfig config,
      Optional<GmmSet> gmms) throws IOException {

    // TODO move this up and pass in tectonic setting?
    // String sourceDirName = dir.getFileName().toString();
    // TectonicSetting tectonicSetting = parseTectonicSetting(sourceDirName);

    System.out.println("loadSourceDirectory(): " + dir.getFileName());

    /* Load GMM tree. */
    Path gmmTree = dir.resolve(GMM_TREE);
    if (gmms.isEmpty() && Files.notExists(gmmTree)) {
      throw new NoSuchFileException(gmmTree.toString());
    }
    GmmSet gmmSet = Files.exists(gmmTree)
        ? readGmms(gmmTree)
        : gmms.get();

    /* Load overriding source config, if present. */
    Path configPath = dir.resolve(SourceConfig.FILE_NAME);
    if (Files.exists(configPath)) {
      config = SourceConfig.fromFile(configPath).build();
      log.info("(override) " + config.toString());
    }

    /*
     * (1) If source tree exists, it governs any subsequent parsing and
     * terminates any further recursion.
     */
    Path sourceTree = dir.resolve(SOURCE_TREE);
    if (Files.exists(sourceTree)) {
      loadSourceTree(sourceTree, builder, config, gmmSet);
      return;
    }

    /* (2) Process any GeoJSON source files. */
    try (DirectoryStream<Path> sourceStream = Files.newDirectoryStream(dir, "*.geojson")) {
      List<Path> sourcePaths = Lists.newArrayList(sourceStream);
      for (Path path : sourcePaths) {
        loadSource(path, builder, config, gmmSet);
      }
    }

    /* (3) Lastly, recurse into any nested directories. */
    try (DirectoryStream<Path> nestedDirStream =
        Files.newDirectoryStream(dir, NestedDirFilter.INSTANCE)) {
      List<Path> nestedDirPaths = Lists.newArrayList(nestedDirStream);
      for (Path path : nestedDirPaths) {
        loadSourceDirectory(path, builder, config, Optional.of(gmmSet));
      }
    }
  }

  /* Process a top-level source-tree. */
  private static void loadSourceTree(
      Path json,
      Builder builder,
      SourceConfig config,
      GmmSet gmmSet) throws IOException {

    System.out.println("     loadSourceTree(): " + json);

    LogicTree<Path> tree = Deserialize.sourceTree(json);
    SourceTree.Builder treeBuilder = SourceTree.builder(tree);

    /* Top-level source trees require a fault-sections directory. */
    treeBuilder.faultSections(readFaultSections(json));

    /* Top-level source tree directories may include a mfd-map. */
    readMfdMap(json).ifPresent(treeBuilder::mfdMap);

    /* Top-level source tree directory may include a mfd-config. */
    MfdConfig mfdConfig = readMfdConfig(json).orElse(MfdConfig.empty());

    for (Branch<Path> branch : tree) {
      /*
       * Must explicitly set mfd-config when descending into branches so that
       * any overrides lower in the tree are reset to the parent/root config
       * upon moving to the next adjacent branch.
       */
      treeBuilder.mfdConfig(mfdConfig);
      processBranch(branch, treeBuilder);
    }

    SourceTree source = treeBuilder.build();
  }

  /* Process source tree branches. */
  private static void processBranch(
      Branch<Path> branch,
      SourceTree.Builder builder) throws IOException {

    System.out.println("      processBranch(): " + branch.value());
    Path tree = branch.value().resolve(SOURCE_TREE);
    // System.out.println("CHILD: " + tree);

    if (Files.exists(tree)) {
      LogicTree<Path> children = Deserialize.sourceTree(tree);
      builder.addBranches(branch, children);
      Optional<MfdConfig> mfdConfig = readMfdConfig(tree);
      for (Branch<Path> child : children) {
        mfdConfig.ifPresent(builder::mfdConfig);
        processBranch(child, builder);
      }
    } else {
      Path ruptureMapPath = branch.value().resolve(RUPTURE_SET);
      // Map<String, NshmRuptureSet> ruptureMap =
      // Deserialize.ruptureMap(ruptureMapPath);
      RuptureSet ruptureSet = Deserialize.ruptureSet(ruptureMapPath, builder.mfdMap);
      Optional<MfdConfig> mfdConfig = readMfdConfig(ruptureMapPath);
      mfdConfig.ifPresent(builder::mfdConfig);

      // TODO clean at this point sourceTreeBuidler should be carrying
      // an MfdCOnfig, rate-tree??
      // System.out.println(ruptureMap.keySet());
      // TODO we are assuming for now that a rate-tree.json, if present,
      // will always be adjacent to a rupture-map.json and so we'll add
      Optional<LogicTree<Double>> rateTree = readRateTree(ruptureMapPath);

      builder.addLeaf(branch, ruptureSet, null);
    }
  }

  /* from source.geojson */
  private static void loadSource(
      Path source,
      Builder builder,
      SourceConfig config,
      GmmSet gmmSet) {

  }

  /*
   * Read ...
   */
  private static GmmSet readGmms(Path json) throws IOException {
    LogicTree<Gmm> tree = Deserialize.gmmTree(json);
    GmmConfig config = Deserialize.gmmConfig(json.resolveSibling(GMM_CONFIG));
    return new GmmSet.Builder().primaryModels(tree, config).build();
  }

  /* Read required fault section map adjacent to top-level source-tree. */
  private static Map<Integer, NshmFaultSection> readFaultSections(Path json)
      throws IOException {
    Path sectionsDir = json.resolveSibling(SECTION_DIR);
    return Deserialize.faultSections(sectionsDir);
  }

  /* Read an optional mfd-map of MFD logic trees. */
  private static Optional<Map<String, LogicTree<MfdProperties>>> readMfdMap(Path sourceTreeJson)
      throws IOException {
    Path mfdMapPath = sourceTreeJson.resolveSibling(MFD_MAP);
    if (Files.exists(mfdMapPath)) {
      return Optional.of(Deserialize.mfdMap(mfdMapPath));
    }
    return Optional.empty();
  }

  /*
   * Read and return an mfd-config, if present, otherwise return
   * MfdConfig.empty(). A MfdConfig is always required when building the
   * branches of a source tree. We do not use Optional.empty() and require
   * explicitely disabling MFD configuration settings using overriding
   * mfd-config.json files deeper in the tree. This permits the inheritance of
   * configuration settings from parent directories; we do not want an absent
   * mfd-config.json file to imply disabling of all MFD configuration settings.
   */
  private static MfdConfig readRootMfdConfig(Path json)
      throws IOException {
    Path mfdConfigPath = json.resolveSibling(MFD_CONFIG);
    MfdConfig mfdConfig = MfdConfig.empty();
    if (Files.exists(mfdConfigPath)) {
      mfdConfig = Deserialize.mfdConfig(mfdConfigPath);
    }
    return mfdConfig;
  }

  /* Read an optional mfd-config. */
  private static Optional<MfdConfig> readMfdConfig(Path json) throws IOException {
    Path mfdConfigPath = json.resolveSibling(MFD_CONFIG);
    if (Files.exists(mfdConfigPath)) {
      return Optional.of(Deserialize.mfdConfig(mfdConfigPath));
    }
    return Optional.empty();
  }

  /* Read an optional rate-tree. */
  private static Optional<LogicTree<Double>> readRateTree(Path json) throws IOException {
    Path rateTreePath = json.resolveSibling(RATE_TREE);
    if (Files.exists(rateTreePath)) {
      return Optional.of(Deserialize.rateTree(rateTreePath));
    }
    return Optional.empty();
  }

  // private static void processTypeDir(
  // Path typeDir,
  // Builder builder,
  // ModelConfig modelConfig,
  // SAXParser sax) throws IOException, SAXException {
  //
  // String typeName = typeDir.getFileName().toString();
  // SourceType type = SourceType.fromString(typeName);
  //
  // /*
  // * gmm.xml file -- this MUST exist if there is at least one source file,
  // * however it MAY NOT exsist if all source files happen to be in nested
  // * directories along with their own gmm.xml files
  // */
  //
  // // Collect type paths
  // List<Path> typePaths = null;
  // try (DirectoryStream<Path> ds = Files.newDirectoryStream(typeDir,
  // SourceFilter.INSTANCE)) {
  // typePaths = Lists.newArrayList(ds);
  // }
  //
  // // load alternate config if such exists
  // ModelConfig config = modelConfig;
  // Path configPath = typeDir.resolve(ModelConfig.FILE_NAME);
  // if (Files.exists(configPath)) {
  // config = ModelConfig.copyOf(modelConfig)
  // .extend(ModelConfig.fromFile(configPath))
  // .build();
  // log.info("(override) " + config.toString());
  // }
  //
  // // Build GmmSet from gmm.xml
  // GmmSet gmmSet = null;
  // Path gmmPath = typeDir.resolve(GmmParser.FILE_NAME);
  //
  // // if source files exist, gmm.xml must also exist
  // if (typePaths.size() > 0 && !Files.exists(gmmPath)) {
  // String mssg = gmmErrorMessage(typeDir.getFileName());
  // throw new FileNotFoundException(mssg);
  // }
  //
  // // having checked directory state, load gmms if present
  // // we may have gmm.xml but no source files
  // if (Files.exists(gmmPath)) {
  // log.info("Parsing: " + typeDir.getParent().relativize(gmmPath));
  // gmmSet = parseGMM(gmmPath, sax);
  // }
  //
  // for (Path sourcePath : typePaths) {
  // log.info("Parsing: " + typeDir.getParent().relativize(sourcePath));
  // SourceSet<? extends Source> sourceSet = parseSource(type, sourcePath,
  // gmmSet, config, sax);
  // builder.sourceSet(sourceSet);
  // }
  //
  // try (DirectoryStream<Path> ds =
  // Files.newDirectoryStream(typeDir, NestedDirFilter.INSTANCE)) {
  // boolean firstDir = true;
  // for (Path nestedSourceDir : ds) {
  // if (firstDir) {
  // log.info("");
  // log.info("======== Nested " + typeName + " Sources ========");
  // firstDir = false;
  // }
  // processNestedDir(nestedSourceDir, type, gmmSet, builder, config, sax);
  // }
  // }
  // }

  // private static void processNestedDir(
  // Path sourceDir,
  // SourceType type,
  // GmmSet gmmSet,
  // Builder builder,
  // ModelConfig parentConfig,
  // SAXParser sax) throws IOException, SAXException {
  //
  // /*
  // * gmm.xml -- this MUST exist if there is at least one source file and there
  // * is no gmm.xml file in the parent source type directory
  // */
  //
  // // Collect nested paths
  // List<Path> nestedSourcePaths = null;
  // try (DirectoryStream<Path> ds = Files.newDirectoryStream(sourceDir,
  // SourceFilter.INSTANCE)) {
  // nestedSourcePaths = Lists.newArrayList(ds);
  // }
  //
  // Path typeDir = sourceDir.getParent().getParent();
  //
  // GmmSet nestedGmmSet = null;
  // ModelConfig nestedConfig = parentConfig;
  // if (nestedSourcePaths.size() > 0) {
  //
  // // config
  // Path nestedConfigPath = sourceDir.resolve(ModelConfig.FILE_NAME);
  // if (Files.exists(nestedConfigPath)) {
  // nestedConfig = ModelConfig.copyOf(parentConfig)
  // .extend(ModelConfig.fromFile(nestedConfigPath))
  // .build();
  // log.info("(override) " + nestedConfig.toString());
  // }
  //
  // // gmm
  // Path nestedGmmPath = sourceDir.resolve(GmmParser.FILE_NAME);
  // if (!Files.exists(nestedGmmPath) && gmmSet == null) {
  // String mssg = gmmErrorMessage(sourceDir.getFileName());
  // throw new FileNotFoundException(mssg);
  // }
  //
  // if (Files.exists(nestedGmmPath)) {
  // log.info("Parsing: " + typeDir.relativize(nestedGmmPath));
  // nestedGmmSet = parseGMM(nestedGmmPath, sax);
  // } else {
  // log.info("(using parent gmm.xml)");
  // nestedGmmSet = gmmSet;
  // }
  //
  // }
  //
  // if (type == SourceType.SYSTEM) {
  // log.info("Parsing: " + typeDir.relativize(sourceDir));
  // parseSystemSource(sourceDir, nestedGmmSet, builder, nestedConfig, sax);
  // } else {
  // for (Path sourcePath : nestedSourcePaths) {
  // log.info("Parsing: " + typeDir.relativize(sourcePath));
  // SourceSet<? extends Source> sourceSet = parseSource(type, sourcePath,
  // nestedGmmSet,
  // nestedConfig, sax);
  // builder.sourceSet(sourceSet);
  // }
  // }
  // }

  // private static SourceSet<? extends Source> parseSource(
  // SourceType type,
  // Path path,
  // GmmSet gmmSet,
  // ModelConfig config,
  // SAXParser sax) throws IOException, SAXException {
  //
  // try (InputStream in = Files.newInputStream(path)) {
  // switch (type) {
  // case AREA:
  // return AreaParser.create(sax).parse(in, gmmSet, config);
  // case CLUSTER:
  // return ClusterParser.create(sax).parse(in, gmmSet, config);
  // case FAULT:
  // return FaultParser.create(sax).parse(in, gmmSet, config);
  // case GRID:
  // return GridParser.create(sax).parse(in, gmmSet, config);
  // case INTERFACE:
  // return InterfaceParser.create(sax).parse(in, gmmSet, config);
  // case SLAB:
  // return SlabParser.create(sax).parse(in, gmmSet, config);
  // case SYSTEM:
  // throw new UnsupportedOperationException(
  // "Fault system sources are not processed with this method");
  // default:
  // throw new IllegalStateException("Unkown source type");
  // }
  // } catch (Exception e) {
  // handleParseException(e, path);
  // return null;
  // }
  // }

  // private static void parseSystemSource(
  // Path dir,
  // GmmSet gmmSet,
  // Builder builder,
  // ModelConfig config,
  // SAXParser sax) throws IOException, SAXException {
  //
  // log.info("");
  // try {
  // Path sectionsPath = dir.resolve(SECTIONS_FILENAME);
  // Path rupturesPath = dir.resolve(RUPTURES_FILENAME);
  // if (Files.exists(sectionsPath) && Files.exists(rupturesPath)) {
  // InputStream sectionsIn = Files.newInputStream(sectionsPath);
  // InputStream rupturesIn = Files.newInputStream(rupturesPath);
  //
  // SystemParser systemParser = SystemParser.create(sax);
  // builder.sourceSet(systemParser.parse(sectionsIn, rupturesIn, gmmSet));
  // } else {
  // log.info("Fault model: (no fault sources supplied with system)");
  // }
  //
  // log.info("");
  // Path gridSourcePath = dir.resolve(GRIDSOURCE_FILENAME);
  // if (Files.exists(gridSourcePath)) {
  // InputStream gridIn = Files.newInputStream(gridSourcePath);
  // GridSourceSet gridSet = GridParser.create(sax).parse(gridIn, gmmSet,
  // config);
  // builder.sourceSet(gridSet);
  // log.info(" Grid model: " + dir.getFileName() + "/" + GRIDSOURCE_FILENAME);
  // log.info(" Weight: " + gridSet.weight());
  // log.info(" Sources: " + gridSet.size());
  // } else {
  // log.info(" Grid model: (no grid sources supplied with system)");
  // }
  // } catch (Exception e) {
  // handleParseException(e, dir);
  // }
  // }

  // private static GmmSet parseGMM(Path path, SAXParser sax)
  // throws IOException, SAXException {
  // try {
  // InputStream in = Files.newInputStream(path);
  // return GmmParser.create(sax).parse(in);
  // } catch (Exception e) {
  // handleParseException(e, path);
  // return null;
  // }
  // }

  // /* Called from all *parse*() methods. */
  // private static void handleParseException(Exception e, Path path)
  // throws IOException, SAXException {
  //
  // StringBuilder sb = new StringBuilder(NEWLINE);
  // sb.append("** Parsing error: ");
  // if (e instanceof SAXParseException) {
  // SAXParseException spe = (SAXParseException) e;
  // sb.append("SAX parser **").append(NEWLINE);
  // sb.append("** Path: ").append(path).append(NEWLINE);
  // sb.append("** Line: ").append(spe.getLineNumber());
  // sb.append(" [").append(spe.getColumnNumber()).append("]").append(NEWLINE);
  // sb.append("** Info: ").append(spe.getMessage());
  // log.severe(sb.toString());
  // throw spe;
  // } else if (e instanceof SAXException) {
  // sb.append("Other SAX error **");
  // log.severe(sb.toString());
  // throw (SAXException) e;
  // } else if (e instanceof IOException) {
  // IOException ioe = (IOException) e;
  // sb.append("IO error ** ").append(NEWLINE);
  // sb.append("** Path: ").append(path);
  // log.severe(sb.toString());
  // throw ioe;
  // } else {
  // sb.append("Other error **");
  // log.severe(sb.toString());
  // throw new SAXException(e);
  // }
  // }

  // private static String gmmErrorMessage(Path dir) {
  // return dir + " sources present. Where is gmm.xml?";
  // }

  private static TectonicSetting parseTectonicSetting(String lowerHyphen) {
    return TectonicSetting.valueOf(LOWER_HYPHEN.to(UPPER_UNDERSCORE, lowerHyphen));
  }

  /*
   * Only lists those directories that match the CaseFormat.LOWER_HYPHEN format
   * of a TectonicSetting.
   */
  private static enum TectonicSettingFilter implements DirectoryStream.Filter<Path> {

    INSTANCE;

    @Override
    public boolean accept(Path path) throws IOException {
      try {
        String name = path.getFileName().toString();
        parseTectonicSetting(name);
        return Files.isDirectory(path);
      } catch (IllegalArgumentException iae) {
        return false;
      }
    }
  }

  // /*
  // * Only lists those directories matching a SourceType.
  // */
  // private static enum TypeFilter implements DirectoryStream.Filter<Path> {
  // INSTANCE;
  //
  // @Override
  // public boolean accept(Path path) throws IOException {
  // try {
  // String name = path.getFileName().toString();
  // SourceType.fromString(name);
  // return true;
  // } catch (IllegalArgumentException iae) {
  // return false;
  // }
  // }
  // }

  /*
   * Filters source XML files, skipping hidden files, those that start with a
   * tilde (~), and any gmm.xml files.
   */
  private static enum SourceFilter implements DirectoryStream.Filter<Path> {
    INSTANCE;

    @Override
    public boolean accept(Path path) throws IOException {
      String s = path.getFileName().toString();
      return Files.isRegularFile(path) && !Files.isHidden(path) &&
          s.toLowerCase().endsWith(".xml") && !s.equals(GmmParser.FILE_NAME) &&
          !s.startsWith("~");
    }
  }

  /*
   * Filters nested source directories, skipping hidden directories, those that
   * start with a tilde (~), or that are named 'sources' that may exist in grid,
   * area, or slab type directories.
   */
  private static enum NestedDirFilter implements DirectoryStream.Filter<Path> {
    INSTANCE;

    @Override
    public boolean accept(Path path) throws IOException {
      String s = path.getFileName().toString();
      return Files.isDirectory(path) && !Files.isHidden(path) && !s.startsWith("~") &&
          !s.equals(GridParser.RATE_DIR);
    }
  }

  // try {
  //
  // SAXParser sax = SAXParserFactory.newInstance().newSAXParser();
  //
  // if (!Files.exists(path)) {
  // String mssg = String.format("Specified model does not exist: %s", path);
  // throw new FileNotFoundException(mssg);
  // }
  //
  // // HazardModel.Builder builder = HazardModel.builder();
  // // List<Path> typePaths = null;
  //
  // // ModelConfig modelConfig = ModelConfig.fromFile(path).build();
  // // log.info(modelConfig.toString());
  //
  // CalcConfig calcConfig = CalcConfig.Builder.withDefaults()
  // .extend(CalcConfig.Builder.fromFile(path))
  // .build();
  // builder.config(calcConfig);
  //
  // typePaths = typeDirectoryList(path);
  // checkState(typePaths.size() > 0, "Empty model: %s", path.getFileName());
  // builder.name(modelConfig.name);
  //
  // for (Path typePath : typePaths) {
  // String typeName = typePath.getFileName().toString();
  // log.info("");
  // log.info("======= " + typeName + " Sources =======");
  // processTypeDir(typePath, builder, modelConfig, sax);
  // log.info("==========================" + Strings.repeat("=",
  // typeName.length()));
  // }
  //
  // log.info("");
  // HazardModel model = builder.build();
  // return model;
  //
  // } catch (ParserConfigurationException oe) {
  // log.severe(NEWLINE + "** Error loading model **");
  // throw new IOException(oe);
  // } catch (Exception e) {
  // log.severe(NEWLINE + "** Error loading model **");
  // throw e;
  // }

}
