package gov.usgs.earthquake.nshmp.eq.model;

import static com.google.common.base.CaseFormat.LOWER_HYPHEN;
import static com.google.common.base.CaseFormat.UPPER_UNDERSCORE;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static com.google.common.base.Strings.padEnd;

import java.io.IOException;
import java.io.Reader;
import java.nio.file.Files;
import java.nio.file.Path;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.annotations.SerializedName;

import gov.usgs.earthquake.nshmp.eq.fault.surface.RuptureFloating;

/*
 * Source configuration. No defaults; 'source-config.json' must be supplied in
 * model root and may be overridden deeper in model file tree. Static factory
 * methods 'copyOf' and 'fromFile' return a builder.
 *
 * @author U.S. Geological Survey
 */
final class SourceConfig {

  static final String ID = "source-config";
  static final String FILE_NAME = ID + ".json";
  private static final String STATE_ERROR = "%s %s not set";

  private static final Gson GSON = new GsonBuilder().create();

  final Path resource;

  final double surfaceSpacing;
  final RuptureFloating ruptureFloating;
  final boolean ruptureVariability;
  final PointSourceType pointSourceType;
  final double areaGridSpacing;

  private SourceConfig(
      Path resource,
      double surfaceSpacing,
      RuptureFloating ruptureFloating,
      boolean ruptureVariability,
      PointSourceType pointSourceType,
      double areaGridSpacing) {

    this.resource = resource;
    this.surfaceSpacing = surfaceSpacing;
    this.ruptureFloating = ruptureFloating;
    this.ruptureVariability = ruptureVariability;
    this.pointSourceType = pointSourceType;
    this.areaGridSpacing = areaGridSpacing;
  }

  static Builder copyOf(SourceConfig that) {
    checkNotNull(that);
    Builder b = new Builder();
    b.resource = that.resource;
    b.surfaceSpacing = that.surfaceSpacing;
    b.ruptureFloating = that.ruptureFloating;
    b.ruptureVariability = that.ruptureVariability;
    b.pointSourceType = that.pointSourceType;
    b.areaGridSpacing = that.areaGridSpacing;
    return b;
  }

  // TODO privatize builder

  /* Supply model-config.json */
  static Builder fromFile(Path json) throws IOException {
    try (Reader reader = Files.newBufferedReader(json)) {
      Builder builder = GSON.fromJson(reader, Builder.class);
      builder.resource = json;
      return builder;
    }
  }

  @Override
  public String toString() {
    return new StringBuilder("source-config: ")
        .append(resource.toAbsolutePath().normalize())
        .append(formatEntry(Key.SURFACE_SPACING, surfaceSpacing))
        .append(formatEntry(Key.RUPTURE_FLOATING, ruptureFloating))
        .append(formatEntry(Key.RUPTURE_VARIABILITY, ruptureVariability))
        .append(formatEntry(Key.POINT_SOURCE_TYPE, pointSourceType))
        .append(formatEntry(Key.AREA_GRID_SPACING, areaGridSpacing))
        .toString();
  }

  private enum Key {
    NAME,
    RESOURCE,
    SURFACE_SPACING,
    RUPTURE_FLOATING,
    RUPTURE_VARIABILITY,
    POINT_SOURCE_TYPE,
    AREA_GRID_SPACING;

    private String label;

    private Key() {
      this.label = UPPER_UNDERSCORE.to(LOWER_HYPHEN, name());
    }

    @Override
    public String toString() {
      return label;
    }
  }

  static <E extends Enum<E>> String formatEntry(E key, Object value) {
    return padEnd("\n  ." + key + ':', 25, ' ') + value;
  }

  final static class Builder {

    private boolean built = false;

    private Path resource;

    @SerializedName("surface-spacing")
    private Double surfaceSpacing;
    @SerializedName("rupture-floating")
    private RuptureFloating ruptureFloating;
    @SerializedName("rupture-variability")
    private Boolean ruptureVariability;
    @SerializedName("point-source-type")
    private PointSourceType pointSourceType;
    @SerializedName("area-grid-spacing")
    private Double areaGridSpacing;

    // TODO clean
    // Builder extend(Builder that) {
    // checkNotNull(that);
    // // resource can't be null
    // this.resource = that.resource;
    // if (that.surfaceSpacing != null) {
    // this.surfaceSpacing = that.surfaceSpacing;
    // }
    // if (that.ruptureFloating != null) {
    // this.ruptureFloating = that.ruptureFloating;
    // }
    // if (that.ruptureVariability != null) {
    // this.ruptureVariability = that.ruptureVariability;
    // }
    // if (that.pointSourceType != null) {
    // this.pointSourceType = that.pointSourceType;
    // }
    // if (that.areaGridSpacing != null) {
    // this.areaGridSpacing = that.areaGridSpacing;
    // }
    // return this;
    // }

    private void validateState() {
      checkState(!built, "This %s builder has already been used", ID);
      checkNotNull(resource, STATE_ERROR, ID, Key.RESOURCE);
      checkNotNull(surfaceSpacing, STATE_ERROR, ID, Key.SURFACE_SPACING);
      checkNotNull(ruptureFloating, STATE_ERROR, ID, Key.RUPTURE_FLOATING);
      checkNotNull(ruptureVariability, STATE_ERROR, ID, Key.RUPTURE_VARIABILITY);
      checkNotNull(pointSourceType, STATE_ERROR, ID, Key.POINT_SOURCE_TYPE);
      checkNotNull(areaGridSpacing, STATE_ERROR, ID, Key.AREA_GRID_SPACING);
      built = true;
    }

    SourceConfig build() {
      validateState();
      return new SourceConfig(
          resource, surfaceSpacing, ruptureFloating,
          ruptureVariability, pointSourceType, areaGridSpacing);
    }
  }

}
