package gov.usgs.earthquake.nshmp.eq.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;

import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import com.google.common.graph.EndpointPair;
import com.google.common.graph.Graphs;
import com.google.common.graph.MutableNetwork;
import com.google.common.graph.Network;
import com.google.common.graph.NetworkBuilder;
import com.google.common.graph.Traverser;

import gov.usgs.earthquake.nshmp.eq.model.MfdHelper2.MfdProperties;
import gov.usgs.earthquake.nshmp.internal.NshmFaultSection;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Top level wrapper class used for any source logic trees. For example, UCERF3
 * alternate fault models, New Madrid, Wasatch...
 * 
 * @author pmpowers
 */
public class SourceTree {

  /*
   * Developer notes:
   * 
   * Root --> Node --> Node --> Leaf
   * 
   * We use Guava's graph library to track source model logic trees. The
   * heirarchy of logic trees in a source model is mapped into a Network where
   * nodes are connected with unique edges that are the branches of the
   * component logic trees.
   * 
   * We tried using a regular Graph with Branches representing the nodes, but it
   * was difficult to extract the node paths between root and Leaf endpoints.
   * 
   * I don't hink we need to really deal with the Node class, it's just
   * connectiuve tissue and supports rooot/trunk and leaf
   * 
   * With regards to sampling, we may want to consider holding a reference to
   * the relevant logic tree at each node.
   * 
   * The internal graph (Network) is a directed tree; each node has only one
   * incident edge/node.
   */

  final Map<Integer, NshmFaultSection> sectionMap;
  final MfdConfig mfdConfig;
  final Optional<Map<String, LogicTree<MfdProperties>>> mfdMap;

  final Network<Node, Branch<Path>> sourceTree;
  final Map<Leaf, List<Branch<Path>>> leafBranches;
  final Map<Leaf, Double> leafWeights;

  private SourceTree(
      Map<Integer, NshmFaultSection> sectionMap,
      MfdConfig mfdConfig,
      Optional<Map<String, LogicTree<MfdProperties>>> mfdMap,
      Network<Node, Branch<Path>> sourceTree,
      Map<Leaf, List<Branch<Path>>> leafBranches,
      Map<Leaf, Double> leafWeights) {

    this.sectionMap = sectionMap;
    this.mfdConfig = mfdConfig;
    this.mfdMap = mfdMap;

    this.sourceTree = sourceTree;
    this.leafBranches = leafBranches;
    this.leafWeights = leafWeights;

    System.out.println(leafWeights);
    System.out.println(leafWeights.values().stream().mapToDouble(d -> d).sum());
    System.out.println(sourceTree);

    for (Leaf leaf : leafWeights.keySet()) {
      System.out.println(leaf.ruptureSet);
      // leaf.ruptureSet.forEach((k, v) -> System.out.println(v.mfdTree));
    }
  }

  /*
   * Return a source tree builder initialized with the trunk or root of the
   * source model logic tree.
   */
  static Builder builder(LogicTree<Path> trunk) {
    return new Builder(trunk);
  }

  static final Root ROOT = new Root();

  static class Builder {

    private int index = 1;

    private Map<Integer, NshmFaultSection> sectionMap;
    private MfdConfig mfdConfig;

    // package access for use during build, required when reading rupture-sets
    Optional<Map<String, LogicTree<MfdProperties>>> mfdMap = Optional.empty();

    private MutableNetwork<Node, Branch<Path>> sourceTree = NetworkBuilder
        .directed()
        .allowsSelfLoops(false)
        .allowsParallelEdges(false)
        .build();

    ImmutableList.Builder<Leaf> leaves = ImmutableList.builder();

    private Builder(LogicTree<Path> tree) {
      for (Branch<Path> branch : tree) {
        sourceTree.addEdge(ROOT, new Node(index++), branch);
      }
    }

    /* Attach the branches of a logic tree to the supplied branch. */
    Builder addBranches(Branch<Path> parent, LogicTree<Path> children) {
      checkArgument(sourceTree.edges().contains(parent), "Missing parent branch");
      EndpointPair<Node> endpoints = sourceTree.incidentNodes(parent);
      // parent target is new source for each branch being added
      Node source = endpoints.target();
      for (Branch<Path> child : children) {
        sourceTree.addEdge(source, new Node(index++), child);
      }
      return this;
    }

    /*
     * Add a leaf to the source tree. This method replaces the target Node
     * associated with the supplied branch with a terminal Leaf node.
     */
    Builder addLeaf(
        Branch<Path> parent,
        RuptureSet ruptureSet,
        Optional<LogicTree<Double>> rateTree) {

      EndpointPair<Node> endpoints = sourceTree.incidentNodes(parent);
      Node target = endpoints.target();
      Node source = endpoints.source();
      sourceTree.removeNode(target);
      Leaf leaf = new Leaf(target.index, ruptureSet); // transfer assigned index
      sourceTree.addEdge(source, leaf, parent);
      leaves.add(leaf);
      return this;
    }

    Builder faultSections(Map<Integer, NshmFaultSection> sectionMap) {
      this.sectionMap = sectionMap;
      return this;
    }

    Builder mfdMap(Map<String, LogicTree<MfdProperties>> mfdMap) {
      this.mfdMap = Optional.of(mfdMap);
      return this;
    }

    /*
     * Because builder descends through source-tree, mfdCnfig must be
     * explicitely reset for each branch.
     */
    Builder mfdConfig(MfdConfig mfdConfig) {
      this.mfdConfig = mfdConfig;
      return this;
    }

    SourceTree build() {
      checkNotNull(sectionMap);
      checkNotNull(mfdConfig);
      List<Leaf> leaves = this.leaves.build();
      Map<Leaf, List<Branch<Path>>> leafBranches = branchLists(sourceTree, leaves);
      Map<Leaf, Double> leafWeights = leafWeights(leafBranches);
      return new SourceTree(
          sectionMap, mfdConfig, mfdMap,
          sourceTree, leafBranches, leafWeights);
    }

    /*
     * Generate lists of logic tree branches (branch paths) from the root of a
     * tree to each of the supplied leaf nodes. Because the tree is a directed
     * graph, traversing a transposed view starting with each leaf yields the
     * required node path.
     */
    private static Map<Leaf, List<Branch<Path>>> branchLists(
        Network<Node, Branch<Path>> tree,
        List<Leaf> leaves) {

      Traverser<Node> traverser = Traverser.forTree(Graphs.transpose(tree));
      ImmutableMap.Builder<Leaf, List<Branch<Path>>> branchListMap = ImmutableMap.builder();
      for (Leaf leaf : leaves) {
        List<Node> nodeList = Lists.newArrayList(traverser.depthFirstPostOrder(leaf));
        checkState(nodeList.size() > 1); // 2 nodes minimum [root -> leaf]
        ImmutableList.Builder<Branch<Path>> branchList = ImmutableList.builder();
        Node source = nodeList.get(0);
        for (Node target : Iterables.skip(nodeList, 1)) {
          branchList.add(tree.edgeConnecting(source, target).get());
          source = target;
        }
        branchListMap.put(leaf, branchList.build());
      }
      return branchListMap.build();
    }

    private static Map<Leaf, Double> leafWeights(Map<Leaf, List<Branch<Path>>> branchLists) {
      return branchLists.entrySet().stream()
          .collect(Collectors.toMap(
              e -> e.getKey(),
              e -> leafWeight(e.getValue())));
    }

    private static double leafWeight(List<Branch<Path>> branchList) {
      return branchList.stream()
          .mapToDouble(Branch::weight)
          .reduce(1, (a, b) -> a * b);
    }
  }

  static class Node {

    int index;

    Node(int index) {
      this.index = index;
    }

    @Override
    public String toString() {
      return "node:" + index;
    }
  }

  static class Root extends Node {

    Root() {
      super(0);
    }

    @Override
    public String toString() {
      return "root:" + index;
    }
  }

  static class Leaf extends Node {

    RuptureSet ruptureSet;

    Leaf(int index, RuptureSet ruptureSet) {
      super(index);
      this.ruptureSet = ruptureSet;
    }

    @Override
    public String toString() {
      return "leaf:" + index;
    }

    RuptureSet ruptureSet() {
      return ruptureSet;
    }
  }
}
