package gov.usgs.earthquake.nshmp.gmm;

import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.sqrt;

import java.util.Set;

import com.google.common.collect.Iterables;
import com.google.common.collect.Streams;

import gov.usgs.earthquake.nshmp.data.Interpolator;

/**
 * BETA implementation<p>
 * 
 * Implement support for Conditional PGV for crustal-earthquake GMM without
 * native PGV support.
 * 
 * <p><b>Implementation notes:</b><ul>
 * 
 * <li>Conditional PGV model for vertical component not yet implemented</li>
 * 
 * <li>PGA-based and PSA(T=1)-based models for PGV (for GMMs with sparse
 * available IMTs) not yet implemented</li> </ul>
 * 
 * @author Peter Powers
 * @author Jason Altekruse
 *
 */
class UsgsPgvSupport {

  // TODO does conditional model behave for different site classes

  /*
   * Coefficients are fixed and not keyed to period so we can include them as
   * static final fields in this class.
   */

  // Table 2: Coefficients for Magnitude Dependence of Tpgv
  private static final double b1 = -4.09;
  private static final double b2 = 0.66;
  // private static final double sigmaTpgv = 0.2;

  // Table 3: Conditional PGV Model Coefficients for Crustal Earthquakes
  // Additional available models: Vertical component, PGA-based, PSA(T=1)-based
  // Horizontal component model:
  private static final double a1 = 5.39;
  private static final double a2 = 0.479;
  private static final double a3 = 0.654;
  private static final double a4 = -0.359;
  private static final double a5 = 0.799;
  private static final double a6 = -0.061;
  // private static final double phi = 0.29;
  // private static final double tau = 0.16;
  private static final double σPgvCond = 0.33;

  private UsgsPgvSupport() {}

  /**
   * Calculates conditional PGV for the specified GMM
   * 
   * @param gmm of type Gmm
   * @param in of type GmmInput
   * @return PGV ScalarGroundMotion
   */
  static ScalarGroundMotion calcPgv(Gmm gmm, GmmInput in) {

    // calculate SA(TPgv), sigma(TPgv)
    ScalarGroundMotion gmTPgv = calcSaGroundMotion(gmm, in);

    // process ground motion according to model

    // Eq. 13: f1 term
    double f1;
    if (in.Mw < 5.0) {
      f1 = a3;
    } else if (in.Mw > 7.5) {
      f1 = a5;
    } else {
      f1 = a3 + (a5 - a3) / 2.5;
    }

    // Eq. 12:
    // there is a typo in the equation - an extra closing parentheses in the a6
    // term
    double lnPgv = a1 + a2 * (in.Mw - 6.0) + a6 * Math.pow(8.5 - in.Mw, 2.0)
    // + f1 * log(gmTPgv.mean())
        + f1 * gmTPgv.mean() + a4 * log(in.rRup + 5.0 * exp(0.4 * (in.Mw - 6.0)));
    double pgv = exp(lnPgv);

    // Eq. 15: aleatory variability
    double σPgv = sqrt(f1 * f1 * gmTPgv.sigma() * gmTPgv.sigma() + σPgvCond * σPgvCond);

    // debug dump: Mw, rRup, TPgv, gmTPgv_mean, gmTPgv_sigma, f1, lnPgv, pgv,
    // sigmaPgv
    // System.out.printf("%.2f\t%.1f\t%.6f\t%.6e\t%.6f\t%.6f\t%.6f\t%.6e\t%.6f\n",
    // in.Mw, in.rRup, computeTPgv(in.Mw), gmTPgv.mean(), gmTPgv.sigma(), f1,
    // lnPgv, pgv, σPgv);

    return DefaultScalarGroundMotion.create(lnPgv, σPgv);
  }

  /*
   * Internal calculation of the SA ground motion for the specified GMM for
   * input to the conditional PGV model
   * 
   * @param Gmm gmm
   * 
   * @param GmmInput in
   * 
   * @return ScalarGroundMotion for target T(PGV)
   */
  private static ScalarGroundMotion calcSaGroundMotion(Gmm gmm, GmmInput in) {
    Set<Imt> saImts = gmm.responseSpectrumIMTs();
    // need to verify that this Set is being returned with ascending
    // iteration order; that is, when we loop over them
    //
    // Iteration order is controlled by set1 in the call to
    // "Sets.intersection(set1, set2)" in gmm.responseSpectrumIMTs()
    // --> If we reverse the order of set1, set2 in that call, the returned
    // set saImts will be sorted in the same order as the IMT Enums
    //
    // Other places responseSpectrumIMTs() is called:
    // HazMatImpl and ResponseSpectra, both in nshmp-haz-v2

    double targetPeriod = computeTPgv(in.Mw);

    // GMM sa Imt bounds
    Imt minImt = saImts.stream().findFirst().get();
    Imt maxImt = Streams.findLast(saImts.stream()).get();
    double minPeriod = minImt.period();
    double maxPeriod = maxImt.period();

    // Check that target is between min max
    // checkState(expression, errorMessage);

    // sa Imt interpolation bounds
    // need periods for interpolation
    // need Imts to compute ground motion
    Imt lowerImt = minImt;
    Imt upperImt = minImt;
    double lowerPeriod = minPeriod; // 0.01s SA known minimum
    double upperPeriod = minPeriod;

    // TODO: throw error if TargetPeriod is outside of the range of lowerPeriod
    // - upperPeriod supported by gmm

    // having set the lower bound with the first element,
    // skip it when looping
    for (Imt saImt : Iterables.skip(saImts, 1)) {
      upperImt = saImt;
      upperPeriod = saImt.period();
      if (upperPeriod > targetPeriod) {
        break;
      }
      lowerImt = upperImt;
      lowerPeriod = upperPeriod;
    }

    /*
     * new InterpolatedGmm(...) currently requires the target to be an IMT, so
     * using this here would require an alternate implementation of the
     * InterpolatedGmm constructor...
     * 
     * Also, we may automatically calculate the spectra at all available IMTs
     * for the GMM up front, so we may just be able to get those values and
     * interpolate without new calls to calc()
     */
    ScalarGroundMotion lowerGm = gmm.instance(lowerImt).calc(in);
    ScalarGroundMotion upperGm = gmm.instance(upperImt).calc(in);

    /*
     * Use the static interpolator method findY. I'm not sure yet if we want to
     * use log(groundMotion), TBD. --> The paper suggests that "the value of
     * PSA(TPgv) should be interpolated using log-log interpolation on the
     * available PSA(T) values"
     */
    double μ = Interpolator.findY(
        log(lowerPeriod), lowerGm.mean(),
        log(upperPeriod), upperGm.mean(),
        log(targetPeriod));

    double σ = Interpolator.findY(
        log(lowerPeriod), lowerGm.sigma(),
        log(upperPeriod), upperGm.sigma(),
        log(targetPeriod));

    return DefaultScalarGroundMotion.create(μ, σ);
  }

  /*
   * Check and document here the range of periods that are spanned by Mw 4.5 to
   * 8.5; I haven't read the paper in detail; I did see 8.5 mentioned as a
   * reasonable upper limit in Mw, but I don't know what the recommended lower
   * limit is.
   *
   * - Recommended upper limit on Mw is 8.5 - Lower limit on Mw is unspecified,
   * but f1 term is conditionally defined for M < 5, and coefficients for
   * PGA-based and PSA(T=1)-based models for PGV may be specified for M>5 but
   * there is a typo in the paper here
   * 
   * 
   * Also check some higher Mw values (that we have in the model) so that we can
   * make some determinations, if necessary, about how to enforce a valid range
   * of magnitudes. I asume that for the range supported by the model that the
   * resultant periods will always be within 0.01 and 10s.
   * 
   * - TPgv ranges from 0.017 to 8.85 sec for M = 0.0 to 9.5, respectively.
   */
  private static double computeTPgv(double M) {
    // Eq. 11
    return exp(b1 + b2 * M);
  }

}
