package gov.usgs.earthquake.nshmp.internal;

import static com.google.common.base.CaseFormat.LOWER_CAMEL;
import static com.google.common.base.CaseFormat.UPPER_UNDERSCORE;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.Earthquakes.INTERFACE_DEPTH_RANGE;
import static gov.usgs.earthquake.nshmp.Faults.DIP_RANGE;
import static gov.usgs.earthquake.nshmp.Faults.RAKE_RANGE;
import static gov.usgs.earthquake.nshmp.data.DoubleData.checkInRange;

import java.net.URL;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson.Type;
import gov.usgs.earthquake.nshmp.geo.json.Properties;

/**
 * Create a new fault section.
 * 
 * @see Builder
 * @author Brandon Clayton
 *
 */
public class NshmFaultSection {

  private static final String MFD_TREE_PROP = "mfd-tree";

  private final Integer id;
  private final String name;
  private final Double rake;
  private final String state;
  private final List<LocationList> traces;

  private final Optional<String> altId;
  private final Optional<String> altIdName;
  private final Optional<Double> dip;
  private final Optional<Double> lowerDepth;
  private final Optional<Object> mfdTree;
  private final Optional<Double> rate;
  private final Optional<RateType> rateType;
  private final Optional<String> secondaryState;
  private final Optional<Double> upperDepth;

  private NshmFaultSection(Builder builder) {
    id = builder.id;
    name = builder.name;
    rake = builder.rake;
    state = builder.state;
    traces = builder.traces;

    altId = Optional.ofNullable(builder.altId);
    altIdName = Optional.ofNullable(builder.altIdName);
    dip = Optional.ofNullable(builder.dip);
    lowerDepth = Optional.ofNullable(builder.lowerDepth);
    mfdTree = Optional.ofNullable(builder.mfdTree);
    rate = Optional.ofNullable(builder.rate);
    rateType = Optional.ofNullable(builder.rateType);
    secondaryState = Optional.ofNullable(builder.secondaryState);
    upperDepth = Optional.ofNullable(builder.upperDepth);
  }

  /**
   * Returns a {@code NshmFaultSection} from a {@code GeoJson} file with a
   * {@code Feature}.
   * 
   * @param json The GeoJson file
   */
  public static NshmFaultSection fromGeoJson(Path json) {
    Feature feature = GeoJson.from(json).toFeature();
    return fromFeature(feature);
  }

  /**
   * Returns a {@code NshmFaultSection} from a {@code GeoJson} file with a
   * {@code Feature}.
   * 
   * @param json The URL to the json file
   */
  public static NshmFaultSection fromGeoJson(URL json) {
    Feature feature = GeoJson.from(json).toFeature();
    return fromFeature(feature);
  }

  /**
   * Return the alternate id.
   */
  public Optional<String> altId() {
    return altId;
  }

  /**
   * Return the alternate id.
   */
  public Optional<String> altIdName() {
    return altIdName;
  }

  /**
   * Returns the fault section dip
   */
  public Optional<Double> dip() {
    return dip;
  }

  /**
   * Returns the fault section id.
   */
  public Integer id() {
    return id;
  }

  /**
   * Returns the fault section lower depth.
   * 
   * <p> Note: lowerDepth >= upperDepth
   */
  public Optional<Double> lowerDepth() {
    return lowerDepth;
  }

  /**
   * Returns the mfd-tree
   */
  public Optional<Object> mfdTree() {
    return mfdTree;
  }

  /**
   * Returns the fault section name.
   */
  public String name() {
    return name;
  }

  /**
   * Returns the fault section rake
   */
  public Double rake() {
    return rake;
  }

  /**
   * Returns the fault section rate
   */
  public Optional<Double> rate() {
    return rate;
  }

  /**
   * Returns the fault section {@code RateType}.
   */
  public Optional<RateType> rateType() {
    return rateType;
  }

  /**
   * Returns the secondary state
   */
  public Optional<String> secondaryState() {
    return secondaryState;
  }

  /**
   * Returns the state were the fault section is.
   * 
   * <p> Note: For arrays, the first state is considered primary
   */
  public String state() {
    return state;
  }

  /**
   * Returns the fault section trace.
   */
  public List<LocationList> traces() {
    return traces;
  }

  /**
   * Returns the fault section upper depth
   */
  public Optional<Double> upperDepth() {
    return upperDepth;
  }

  /**
   * Returns a {@code Feature}.
   */
  public Feature toFeature() {
    Feature.Builder builder = null;
    List<LocationList> traces = traces();

    if (traces.size() == 1) {
      builder = Feature.lineString(traces().get(0));
    } else {
      builder = Feature.multiLineString(traces());
    }

    return builder.id(id())
        .properties(setProperties())
        .build();
  }

  private Map<String, Object> setProperties() {
    Properties.Builder properties = Properties.builder();

    properties.put(PropertyKey.NAME.toString(), name());

    if (secondaryState.isPresent()) {
      properties.put(
          PropertyKey.STATE.toString(),
          new String[] { state(), secondaryState().get() });
    } else {
      properties.put(PropertyKey.STATE.toString(), state());
    }

    if (altId.isPresent() && altIdName.isPresent()) {
      properties.put(altIdName().get(), altId().get());
    }

    if (dip.isPresent()) {
      properties.put(PropertyKey.DIP.toString(), dip().get());
    }

    if (lowerDepth.isPresent()) {
      properties.put(PropertyKey.LOWER_DEPTH.toString(), lowerDepth().get());
    }

    properties.put(PropertyKey.RAKE.toString(), rake());

    if (rate.isPresent()) {
      properties.put(PropertyKey.RATE.toString(), rate().get());
    }

    if (rateType.isPresent()) {
      properties.put(PropertyKey.RATE_TYPE.toString(), rateType().get());
    }

    if (upperDepth.isPresent()) {
      properties.put(PropertyKey.UPPER_DEPTH.toString(), upperDepth().get());
    }

    return properties.build();
  }

  /**
   * Returns the {@code FaultSection} {@code Builder}
   * 
   * @see Builder
   */
  public static Builder builder() {
    return new Builder();
  }

  /**
   * Build a {@code FaultSection}. <br><br>
   * 
   * Note: {@link #name(String)} and {@link #state(String...)} must be set.
   */
  public static class Builder {
    private String altId;
    private String altIdName;
    private Double dip;
    private Integer id;
    private Double lowerDepth;
    private Object mfdTree;
    private String name;
    private Double rake;
    private Double rate;
    private RateType rateType;
    private String secondaryState;
    private String state;
    private List<LocationList> traces;
    private Double upperDepth;

    private Builder() {
      altId = null;
      altIdName = null;
      dip = null;
      id = null;
      lowerDepth = null;
      mfdTree = null;
      name = null;
      rake = null;
      rate = null;
      secondaryState = null;
      state = null;
      upperDepth = null;
    }

    /**
     * Set an optional alternate id.
     * @param altIdName The alternate id name
     * @param altId The alternate id
     * @return The builder
     */
    public Builder altId(String altIdName, String altId) {
      this.altId = altId;
      this.altIdName = altIdName;
      return this;
    }

    /**
     * Set the optional dip in degrees. [0..90]°
     * 
     * <p> Note: Required for non-interface fault section.
     * 
     * @param dip The dip of the fault section
     * @return The builder
     */
    public Builder dip(Double dip) {
      checkInRange(DIP_RANGE, dip);
      this.dip = dip;
      return this;
    }

    /**
     * Set the fault section id.
     * @param id The fault section id
     * @return The builder
     */
    public Builder id(Integer id) {
      this.id = id;
      return this;
    }

    /**
     * Set the optional lower depth in km (positive down). [0..60] km <br> Note:
     * lowerDepth >= upperDepth.
     * 
     * <p> Note: Required for non-interface fault section.
     * 
     * @param lowerDepth The lower depth of the fault section
     * @return The builder
     */
    public Builder lowerDepth(Double lowerDepth) {
      checkInRange(INTERFACE_DEPTH_RANGE, lowerDepth);
      this.lowerDepth = lowerDepth;
      return this;
    }

    /**
     * Set the optional mfd tree.
     * @param mfdTree The mfd tree
     * @return The builder
     */
    public Builder mfdTree(Object mfdTree) {
      this.mfdTree = mfdTree;
      return this;
    }

    /**
     * Set the name.
     * @param name The name of the fault section
     * @return The builder
     */
    public Builder name(String name) {
      this.name = name;
      return this;
    }

    /**
     * Set the rake in degrees. [-180..180]°
     * @param rake The rake of the fault section
     * @return
     */
    public Builder rake(Double rake) {
      checkInRange(RAKE_RANGE, rake);
      this.rake = rake;
      return this;
    }

    /**
     * Set the optional rate in mm/yr.
     * @param rate The slip or displacement rate of the fault section
     * @return The builder
     */
    public Builder rate(Double rate) {
      this.rate = rate;
      return this;
    }

    /**
     * Set the optional {@code RateType}.
     * @param rateType The rate type of the fault section
     * @return The builder
     */
    public Builder rateType(RateType rateType) {
      this.rateType = rateType;
      return this;
    }

    /**
     * Set the optional secondary state.
     * @param secondaryState The secondary state
     * @return The builder
     */
    public Builder secondaryState(String secondaryState) {
      this.secondaryState = secondaryState;
      return this;
    }

    /**
     * Set the state were fault section is located.
     * @param state The state were fault section is
     * @return The builder
     */
    public Builder state(String state) {
      this.state = state;
      return this;
    }

    /**
     * Set the trace.
     * @param trace The trace of the fault section
     * @return The builder
     */
    public Builder traces(List<LocationList> traces) {
      this.traces = traces;
      return this;
    }

    /**
     * Set the optional upper depth in km (positive down). [0..60] km
     * 
     * <p> Note: Required for non-interface fault section.
     * 
     * @param upperDepth
     * @return The builder
     */
    public Builder upperDepth(Double upperDepth) {
      checkInRange(INTERFACE_DEPTH_RANGE, upperDepth);
      this.upperDepth = upperDepth;
      return this;
    }

    /**
     * Returns a new {@code NshmFaultSection}
     */
    public NshmFaultSection toFaultSection() {
      return new NshmFaultSection(validate(false));
    }

    /**
     * Returns a new interface {@code NshmFaultSection}
     */
    public NshmFaultSection toInterfaceFaultSection() {
      return new NshmFaultSection(validate(true));
    }

    private Builder validate(boolean isInterface) {
      checkNotNull(id, "id not set");
      checkNotNull(name, "name not set");
      checkNotNull(rake, "rake not set");
      checkNotNull(state, "state not set");
      checkNotNull(traces, "traces not set");

      // TODO make rate and rateType required for regular fault section
      if (isInterface) {
        checkState(dip == null, "dip cannot be set for interface");
        checkState(lowerDepth == null, "lower depth cannot be set for interface");
        checkState(upperDepth == null, "upper depth cannot be set for interface");
      } else {
        checkNotNull(dip, "dip not set");
        checkNotNull(lowerDepth, "lower depth not set");
        checkNotNull(upperDepth, "upper depth not set");
        checkState(
            lowerDepth >= upperDepth,
            "lowerDepth [%f] < upperDepth [%f]",
            lowerDepth,
            upperDepth);
      }

      return this;
    }

    private void setField(PropertyKey key, Properties properties) {
      Object value = properties.get(key.toString());
      switch (key) {
        case DIP:
          dip((Double) value);
          break;
        case LOWER_DEPTH:
          lowerDepth((Double) value);
          break;
        case MFD_TREE:
          mfdTree(value);
          break;
        case NAME:
          name((String) value);
          break;
        case RAKE:
          rake((Double) value);
          break;
        case RATE:
          rate((Double) value);
          break;
        case RATE_TYPE:
          rateType(RateType.valueOf((String) value));
          break;
        case STATE:
          if (value.getClass().isInstance(new ArrayList<String>())) {
            String[] states = properties.get(key.toString(), String[].class);
            state(states[0]);
            secondaryState(states[1]);
          } else {
            state((String) value);
          }
          break;
        case UPPER_DEPTH:
          upperDepth((Double) value);
          break;
        default:
          throw new RuntimeException("Property [" + key + "] not allowed");
      }
    }
  }

  private static NshmFaultSection fromFeature(Feature feature) {
    Properties properties = feature.properties();
    Type type = feature.type();
    Builder builder = new Builder();
    builder.id(feature.idAsInt());

    for (PropertyKey key : PropertyKey.values()) {
      if (properties.containsKey(key.toString())) {
        builder.setField(key, properties);
      }
    }

    properties.map().forEach((key, value) -> {
      if (key.endsWith("Id")) {
        builder.altId(key, value.isJsonNull() ? null : value.getAsString());
      }
    });

    switch (type) {
      case LINE_STRING:
        List<LocationList> locs = new ArrayList<>();
        locs.add(feature.asLineString());
        builder.traces(locs);
        return builder.toFaultSection();
      case MULTI_LINE_STRING:
        builder.traces(feature.asMultiLineString());
        return builder.toInterfaceFaultSection();
      default:
        throw new IllegalStateException("GeoJson type [" + type + "] not supported");
    }
  }

  static enum PropertyKey {
    DIP,
    LOWER_DEPTH,
    MFD_TREE,
    NAME,
    RAKE,
    RATE,
    RATE_TYPE,
    STATE,
    UPPER_DEPTH;

    @Override
    public String toString() {
      if (name().equals(MFD_TREE.name())) {
        return MFD_TREE_PROP;
      } else {
        return UPPER_UNDERSCORE.to(LOWER_CAMEL, name());
      }
    }
  }

}
