package gov.usgs.earthquake.nshmp.tree;

import static com.google.common.base.Preconditions.checkState;

import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.ImmutableList;

import gov.usgs.earthquake.nshmp.data.DoubleData;

/* Basic logic tree implementation. */
class RegularTree<T> implements LogicTree<T> {

  private final List<Branch<T>> branches;
  private final double[] cumulativeWeights;

  RegularTree(List<Branch<T>> branches) {
    checkState(!branches.isEmpty(), "Empty logic tree");
    this.branches = branches;
    double[] weights = weightArray(branches);
    DoubleData.checkWeightSum(weights);
    // round out double precision errors introduced during cumulation
    cumulativeWeights = DoubleData.round(6, DoubleData.cumulate(weights));
  }

  private static <T> double[] weightArray(List<Branch<T>> branches) {
    return branches.stream()
        .mapToDouble(Branch::weight)
        .toArray();
  }

  @Override
  public Branch<T> sample(double probability) {
    for (int i = 0; i < cumulativeWeights.length; i++) {
      if (probability < cumulativeWeights[i]) {
        return branches.get(i);
      }
    }
    return branches.get(cumulativeWeights.length - 1);
  }

  @Override
  public List<Branch<T>> sample(double[] probabilities) {
    ImmutableList.Builder<Branch<T>> samples =
        ImmutableList.builderWithExpectedSize(probabilities.length);
    for (double probability : probabilities) {
      samples.add(sample(probability));
    }
    return samples.build();
  }

  @Override
  public Iterator<Branch<T>> iterator() {
    return branches.iterator();
  }

  @Override
  public List<Branch<T>> branches() {
    return branches;
  }

  @VisibleForTesting
  double[] cumulativeWeights() {
    return Arrays.copyOf(cumulativeWeights, cumulativeWeights.length);
  }

}
