package gov.usgs.earthquake.nshmp.data;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.util.Arrays;
import java.util.List;

import org.junit.jupiter.api.Test;

import com.google.common.collect.ImmutableList;

class SequencesTests {

  private static final XySequence xyBase;
  private static final XySequence xySameXs;
  private static final XySequence xyDifferentXs;
  private static final XySequence xyShorter;

  private static final List<XySequence> emptyCollection;
  private static final List<XySequence> singleSequence;
  private static final List<XySequence> identicalSequences;
  private static final List<XySequence> sameXs;
  private static final List<XySequence> differentXs;
  private static final List<XySequence> withShorterXs;

  private static final XySequence expectedSingleSequenceResults;
  private static final XySequence expectedIdenticalSequences;
  private static final XySequence expectedSameXs;
  private static final XySequence expectedDifferentXs;
  private static final XySequence expectedWithShorterXs;

  static {
    double[] xs1 = new double[] { 0, 1, 2, 3 };
    double[] xs2 = new double[] { 1, 2, 3, 4 };
    double[] xs3 = new double[] { 5, 6, 7 };

    double[] ys1 = new double[] { 10, 9, 8, 7 };
    double[] ys2 = new double[] { 5, 4, 3, 2 };
    double[] ys3 = new double[] { 10, 9, 8 };

    xyBase = XySequence.create(xs1, ys1);
    xySameXs = XySequence.create(xs1, ys2);
    xyDifferentXs = XySequence.create(xs2, ys2);
    xyShorter = XySequence.create(xs3, ys3);

    emptyCollection = new ImmutableList.Builder<XySequence>().build();

    singleSequence = new ImmutableList.Builder<XySequence>()
        .add(xyBase)
        .build();
    expectedSingleSequenceResults = xyBase;

    // add two copies of the same sequence
    identicalSequences = new ImmutableList.Builder<XySequence>()
        .add(xyBase)
        .add(xyBase)
        .build();
    double[] result = new double[xs1.length];
    Arrays.setAll(result, i -> ys1[i] + ys1[i]);
    expectedIdenticalSequences = XySequence.create(xs1, result);

    // sequences with the same x-values (similar)
    sameXs = new ImmutableList.Builder<XySequence>()
        .add(xyBase)
        .add(xySameXs)
        .build();
    Arrays.setAll(result, i -> ys1[i] + ys2[i]);
    expectedSameXs = XySequence.create(xs1, result);

    // sequences with different x-values
    differentXs = new ImmutableList.Builder<XySequence>()
        .add(xyBase)
        .add(xyDifferentXs)
        .build();
    // manually construct for now
    expectedDifferentXs = XySequence.create(
        new double[] { 0, 1, 2, 3, 4 },
        new double[] { 10, 14, 12, 10, 2 });

    // sequences with different x-values and different length
    withShorterXs = new ImmutableList.Builder<XySequence>()
        .add(xyBase)
        .add(xyDifferentXs)
        .add(xyShorter)
        .build();
    // manually construct for now
    expectedWithShorterXs = XySequence.create(
        new double[] { 0, 1, 2, 3, 4, 5, 6, 7 },
        new double[] { 10, 14, 12, 10, 2, 10, 9, 8 });
  }

  @Test
  final void areSimilarTests_emptyThrowsISE() {
    assertThrows(IllegalStateException.class, () -> {
      Sequences.areSimilar(emptyCollection);
    });
  }

  @Test
  final void areSimilarTests() {
    assertTrue(Sequences.areSimilar(singleSequence));
    assertTrue(Sequences.areSimilar(identicalSequences));
    assertTrue(Sequences.areSimilar(sameXs));
    assertFalse(Sequences.areSimilar(differentXs));
    assertFalse(Sequences.areSimilar(withShorterXs));
  }

  @Test
  final void combineTests_emptyThrowsISE() {
    assertThrows(IllegalStateException.class, () -> {
      Sequences.combine(emptyCollection);
    });
  }

  @Test
  final void combineTests_similarSequences() {
    assertEquals(expectedSingleSequenceResults, Sequences.combine(singleSequence));
    assertEquals(expectedIdenticalSequences, Sequences.combine(identicalSequences));
    assertEquals(expectedSameXs, Sequences.combine(sameXs));
  }

  @Test
  final void combineTests_dissimilarSequences() {
    assertEquals(expectedDifferentXs, Sequences.combine(differentXs));
    assertEquals(expectedWithShorterXs, Sequences.combine(withShorterXs));
  }

}
