package gov.usgs.earthquake.nshmp.internal;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertSame;
import static org.junit.jupiter.api.Assertions.assertThrows;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.OptionalDouble;
import java.util.stream.Collectors;

import org.junit.jupiter.api.Test;

import com.google.common.collect.ImmutableList;

import gov.usgs.earthquake.nshmp.internal.Csv.Record;

@SuppressWarnings("javadoc")
public class CsvTests {

  private static final Path DATA_PATH = Paths.get("src/test/resources/internal");

  private static final Path CSV_TESTS = DATA_PATH.resolve("csv-tests.csv");

  @Test
  public void testCreateExceptions() {
    assertThrows(NullPointerException.class, () -> {
      List<String> lines = null;
      Csv.create(lines);
    });

    assertThrows(NullPointerException.class, () -> {
      Path path = null;
      Csv.create(path);
    });

    assertThrows(IllegalArgumentException.class, () -> {
      Csv.create(DATA_PATH.resolve("no-file.csv"));
    });

    assertThrows(IllegalStateException.class, () -> {
      Csv.create(DATA_PATH.resolve("csv-tests-empty.csv"));
    });

    assertThrows(IllegalStateException.class, () -> {
      Csv.create(DATA_PATH.resolve("csv-tests-bad-header-1.csv"));
    });

    assertThrows(IllegalStateException.class, () -> {
      Csv.create(DATA_PATH.resolve("csv-tests-bad-header-2.csv"));
    });
  }

  private static final List<String> COLUMN_KEYS = ImmutableList.of(
      "lon", "lat", "a", "b", "c", "d");

  @Test
  public void testColumnKeys() throws IOException {
    Csv csvPath = Csv.create(CSV_TESTS);
    Csv csvLines = Csv.create(Files.readAllLines(CSV_TESTS));
    assertEquals(COLUMN_KEYS, csvPath.columnKeys());
    assertEquals(COLUMN_KEYS, csvLines.columnKeys());
    assertEquals(csvPath.columnKeys(), csvLines.columnKeys());
  }

  @Test
  public void testRecords() throws IOException {
    Csv csvPath = Csv.create(CSV_TESTS);
    Csv csvLines = Csv.create(Files.readAllLines(CSV_TESTS));

    List<Record> recordsPath = csvPath.records().collect(Collectors.toList());
    List<Record> recordsLines = csvLines.records().collect(Collectors.toList());
    assertEquals(-100, recordsPath.get(0).getDouble("lon"), 0.0);
    assertEquals(-100, recordsLines.get(0).getDouble("lon"), 0.0);

    assertEquals(2, recordsPath.get(0).getDouble("b"), 0.0);
    assertEquals(2, recordsLines.get(0).getDouble("b"), 0.0);

    assertEquals(false, recordsPath.get(1).getBoolean("d"));
    assertEquals(false, recordsLines.get(1).getBoolean("d"));

    assertEquals("test", recordsPath.get(2).get("c"));
    assertEquals("test", recordsLines.get(2).get("c"));

    assertSame(OptionalDouble.empty(), recordsPath.get(2).getOptionalDouble("a"));
    assertSame(OptionalDouble.empty(), recordsLines.get(2).getOptionalDouble("a"));

    assertEquals(20, recordsPath.get(2).getOptionalDouble("lat").getAsDouble(), 0.0);
    assertEquals(20, recordsLines.get(2).getOptionalDouble("lat").getAsDouble(), 0.0);

    assertThrows(IllegalStateException.class, () -> {
      Csv csv2 = Csv.create(DATA_PATH.resolve("csv-tests-bad-record.csv"));
      csv2.records().collect(Collectors.toList());
    });
  }

  @Test
  public void testToString() throws IOException {
    String expected = "[-110, 30, , , true, false]";
    String actual = Csv.create(DATA_PATH.resolve("csv-tests.csv"))
        .records()
        .skip(1)
        .findFirst()
        .get()
        .toString();
    assertEquals(expected, actual);
  }

}
